@@ -9,12 +9,11 @@ Python 3 socket module.
 from __future__ import absolute_import
 import io
 import os
-import sys
+
 
 from gevent import _socketcommon
 from gevent._util import copy_globals
 from gevent._compat import PYPY
-from gevent.timeout import Timeout
 import _socket
 from os import dup
 
@@ -23,13 +22,6 @@ copy_globals(_socketcommon, globals(),
              names_to_ignore=_socketcommon.__extensions__,
              dunder_names_to_keep=())
 
-try:
-    from errno import EHOSTUNREACH
-    from errno import ECONNREFUSED
-except ImportError:
-    EHOSTUNREACH = -1
-    ECONNREFUSED = -1
-
 
 __socket__ = _socketcommon.__socket__
 __implements__ = _socketcommon._implements
@@ -40,9 +32,6 @@ __dns__ = _socketcommon.__dns__
 
 SocketIO = __socket__.SocketIO # pylint:disable=no-member
 
-from gevent._greenlet_primitives import get_memory as _get_memory
-
-timeout_default = object()
 
 class _closedsocket(object):
     __slots__ = ('family', 'type', 'proto', 'orig_fileno', 'description')
@@ -62,7 +51,7 @@ class _closedsocket(object):
 
     detach = fileno
 
-    def _dummy(*args, **kwargs): # pylint:disable=no-method-argument,unused-argument
+    def _dummy(*args, **kwargs): # pylint:disable=no-method-argument,unused-argument,no-self-argument
         raise OSError(EBADF, 'Bad file descriptor')
     # All _delegate_methods must also be initialized here.
     send = recv = recv_into = sendto = recvfrom = recvfrom_into = _dummy
@@ -96,7 +85,6 @@ class _wrefsocket(_socket.socket):
         timeout = property(lambda s: s.gettimeout(),
                            lambda s, nv: s.settimeout(nv))
 
-from gevent._hub_primitives import wait_on_socket as _wait_on_socket
 
 class socket(_socketcommon.SocketMixin):
     """
@@ -113,37 +101,31 @@ class socket(_socketcommon.SocketMixin):
     # of _wrefsocket. (gevent internal usage only)
     _gevent_sock_class = _wrefsocket
 
-    _io_refs = 0
-    _closed = False
-    _read_event = None
-    _write_event = None
-
+    __slots__ = (
+        '_io_refs',
+        '_closed',
+    )
 
     # Take the same approach as socket2: wrap a real socket object,
     # don't subclass it. This lets code that needs the raw _sock (not tied to the hub)
     # get it. This shows up in tests like test__example_udp_server.
 
-    if sys.version_info[:2] < (3, 7):
-        def __init__(self, family=AF_INET, type=SOCK_STREAM, proto=0, fileno=None):
-            self._sock = self._gevent_sock_class(family, type, proto, fileno)
-            self.timeout = None
-            self.__init_common()
-    else:
-        # In 3.7, socket changed to auto-detecting family, type, and proto
-        # when given a fileno.
-        def __init__(self, family=-1, type=-1, proto=-1, fileno=None):
-            if fileno is None:
-                if family == -1:
-                    family = AF_INET
-                if type == -1:
-                    type = SOCK_STREAM
-                if proto == -1:
-                    proto = 0
-            self._sock = self._gevent_sock_class(family, type, proto, fileno)
-            self.timeout = None
-            self.__init_common()
+    # In 3.7, socket changed to auto-detecting family, type, and proto
+    # when given a fileno.
+    def __init__(self, family=-1, type=-1, proto=-1, fileno=None):
+        super().__init__()
+        self._closed = False
+        if fileno is None:
+            if family == -1:
+                family = AddressFamily.AF_INET
+            if type == -1:
+                type = SOCK_STREAM
+            if proto == -1:
+                proto = 0
+        self._sock = self._gevent_sock_class(family, type, proto, fileno)
+        self.timeout = None
 
-    def __init_common(self):
+        self._io_refs = 0
         _socket.socket.setblocking(self._sock, False)
         fileno = _socket.socket.fileno(self._sock)
         self.hub = get_hub()
@@ -155,6 +137,11 @@ class socket(_socketcommon.SocketMixin):
     def __getattr__(self, name):
         return getattr(self._sock, name)
 
+    def _accept(self):
+        # Python 3.11 started checking for this method on the class object,
+        # so we need to explicitly delegate.
+        return self._sock._accept()
+
     if hasattr(_socket, 'SOCK_NONBLOCK'):
         # Only defined under Linux
         @property
@@ -164,16 +151,6 @@ class socket(_socketcommon.SocketMixin):
                 return self._sock.type & ~_socket.SOCK_NONBLOCK # pylint:disable=no-member
             return self._sock.type
 
-    def getblocking(self):
-        """
-        Returns whether the socket will approximate blocking
-        behaviour.
-
-        .. versionadded:: 1.3a2
-            Added in Python 3.7.
-        """
-        return self.timeout != 0.0
-
     def __enter__(self):
         return self
 
@@ -208,17 +185,6 @@ class socket(_socketcommon.SocketMixin):
     def __getstate__(self):
         raise TypeError("Cannot serialize socket object")
 
-    def _get_ref(self):
-        return self._read_event.ref or self._write_event.ref
-
-    def _set_ref(self, value):
-        self._read_event.ref = value
-        self._write_event.ref = value
-
-    ref = property(_get_ref, _set_ref)
-
-    _wait = _wait_on_socket
-
     def dup(self):
         """dup() -> socket object
 
@@ -311,12 +277,27 @@ class socket(_socketcommon.SocketMixin):
         # so that (hopefully) they can be closed before we destroy
         # the FD and invalidate them. We may be in the hub running pending
         # callbacks now, or this may take until the next iteration.
-        scheduled_new = self.hub.loop.closing_fd(sock.fileno())
+        should_defer = self.hub.loop.closing_fd(sock.fileno())
         # Schedule the actual close to happen after that, but only if needed.
         # (If we always defer, we wind up closing things much later than expected.)
-        if scheduled_new:
-            self.hub.loop.run_callback(sock.close)
+        # Note that if we're in the middle of running callbacks, simply scheduling
+        # a callback with ``run_callback`` could result in the callback being
+        # called IMMEDIATELY, which completely defeats the point.
+        if should_defer:
+            check = self.hub.loop.check()
+            def cb(s):
+                try:
+                    s.close()
+                except OSError:
+                    pass
+                finally:
+                    check.stop()
+                    check.close()
+            check.start(cb, sock)
         else:
+            # Note that if the file descriptor got closed (``os.close``)
+            # closing the socket will now raise OSError: EBADF. The
+            # stdlib does the same.
             sock.close()
 
 
@@ -345,6 +326,7 @@ class socket(_socketcommon.SocketMixin):
         # Break any reference to the loop.io objects. Our fileno,
         # which they were tied to, is about to be free to be reused, so these
         # objects are no longer functional.
+        # pylint:disable-next=superfluous-parens
         self._drop_events_and_close(closefd=(reason == 'closed'))
 
         self._sock = _closedsocket(family, type, proto, fileno, reason)
@@ -388,71 +370,6 @@ class socket(_socketcommon.SocketMixin):
         self._detach_socket('detached')
         return sock.detach()
 
-    def connect(self, address):
-        """
-        Connect to *address*.
-
-        .. versionchanged:: 20.6.0
-            If the host part of the address includes an IPv6 scope ID,
-            it will be used instead of ignored, if the platform supplies
-            :func:`socket.inet_pton`.
-        """
-        if self.timeout == 0.0:
-            return _socket.socket.connect(self._sock, address)
-        address = _socketcommon._resolve_addr(self._sock, address)
-        with Timeout._start_new_or_dummy(self.timeout, timeout("timed out")):
-            while True:
-                err = self.getsockopt(SOL_SOCKET, SO_ERROR)
-                if err:
-                    raise error(err, strerror(err))
-                result = _socket.socket.connect_ex(self._sock, address)
-
-                if not result or result == EISCONN:
-                    break
-                if (result in (EWOULDBLOCK, EINPROGRESS, EALREADY)) or (result == EINVAL and is_windows):
-                    self._wait(self._write_event)
-                else:
-                    if (isinstance(address, tuple)
-                            and address[0] == 'fe80::1'
-                            and result == EHOSTUNREACH):
-                        # On Python 3.7 on mac, we see EHOSTUNREACH
-                        # returned for this link-local address, but it really is
-                        # supposed to be ECONNREFUSED according to the standard library
-                        # tests (test_socket.NetworkConnectionNoServer.test_create_connection)
-                        # (On previous versions, that code passed the '127.0.0.1' IPv4 address, so
-                        # ipv6 link locals were never a factor; 3.7 passes 'localhost'.)
-                        # It is something of a mystery how the stdlib socket code doesn't
-                        # produce EHOSTUNREACH---I (JAM) can't see how socketmodule.c would avoid
-                        # that. The normal connect just calls connect_ex much like we do.
-                        result = ECONNREFUSED
-                    raise error(result, strerror(result))
-
-    def connect_ex(self, address):
-        try:
-            return self.connect(address) or 0
-        except timeout:
-            return EAGAIN
-        except gaierror: # pylint:disable=try-except-raise
-            # gaierror/overflowerror/typerror is not silenced by connect_ex;
-            # gaierror extends OSError (aka error) so catch it first
-            raise
-        except error as ex:
-            # error is now OSError and it has various subclasses.
-            # Only those that apply to actually connecting are silenced by
-            # connect_ex.
-            if ex.errno:
-                return ex.errno
-            raise # pragma: no cover
-
-    def recv(self, *args):
-        while True:
-            try:
-                return self._sock.recv(*args)
-            except error as ex:
-                if ex.args[0] != EWOULDBLOCK or self.timeout == 0.0:
-                    raise
-            self._wait(self._read_event)
-
     if hasattr(_socket.socket, 'recvmsg'):
         # Only on Unix; PyPy 3.5 5.10.0 provides sendmsg and recvmsg, but not
         # recvmsg_into (at least on os x)
@@ -481,72 +398,6 @@ class socket(_socketcommon.SocketMixin):
                         raise
                 self._wait(self._read_event)
 
-    def recvfrom(self, *args):
-        while True:
-            try:
-                return self._sock.recvfrom(*args)
-            except error as ex:
-                if ex.args[0] != EWOULDBLOCK or self.timeout == 0.0:
-                    raise
-            self._wait(self._read_event)
-
-    def recvfrom_into(self, *args):
-        while True:
-            try:
-                return self._sock.recvfrom_into(*args)
-            except error as ex:
-                if ex.args[0] != EWOULDBLOCK or self.timeout == 0.0:
-                    raise
-            self._wait(self._read_event)
-
-    def recv_into(self, *args):
-        while True:
-            try:
-                return self._sock.recv_into(*args)
-            except error as ex:
-                if ex.args[0] != EWOULDBLOCK or self.timeout == 0.0:
-                    raise
-            self._wait(self._read_event)
-
-    def send(self, data, flags=0, timeout=timeout_default):
-        if timeout is timeout_default:
-            timeout = self.timeout
-        try:
-            return self._sock.send(data, flags)
-        except error as ex:
-            if ex.args[0] not in _socketcommon.GSENDAGAIN or timeout == 0.0:
-                raise
-            self._wait(self._write_event)
-            try:
-                return _socket.socket.send(self._sock, data, flags)
-            except error as ex2:
-                if ex2.args[0] == EWOULDBLOCK:
-                    return 0
-                raise
-
-    def sendall(self, data, flags=0):
-        # XXX Now that we run on PyPy3, see the notes in _socket2.py's sendall()
-        # and implement that here if needed.
-        # PyPy3 is not optimized for performance yet, and is known to be slower than
-        # PyPy2, so it's possibly premature to do this. However, there is a 3.5 test case that
-        # possibly exposes this in a severe way.
-        data_memory = _get_memory(data)
-        return _socketcommon._sendall(self, data_memory, flags)
-
-    def sendto(self, *args):
-        try:
-            return self._sock.sendto(*args)
-        except error as ex:
-            if ex.args[0] != EWOULDBLOCK or self.timeout == 0.0:
-                raise
-            self._wait(self._write_event)
-            try:
-                return self._sock.sendto(*args)
-            except error as ex2:
-                if ex2.args[0] == EWOULDBLOCK:
-                    return 0
-                raise
-
     if hasattr(_socket.socket, 'sendmsg'):
         # Only on Unix
         def sendmsg(self, buffers, ancdata=(), flags=0, address=None):
@@ -568,40 +419,6 @@ class socket(_socketcommon.SocketMixin):
                         return 0
                     raise
 
-    def setblocking(self, flag):
-        # Beginning in 3.6.0b3 this is supposed to raise
-        # if the file descriptor is closed, but the test for it
-        # involves closing the fileno directly. Since we
-        # don't touch the fileno here, it doesn't make sense for
-        # us.
-        if flag:
-            self.timeout = None
-        else:
-            self.timeout = 0.0
-
-    def settimeout(self, howlong):
-        if howlong is not None:
-            try:
-                f = howlong.__float__
-            except AttributeError:
-                raise TypeError('a float is required')
-            howlong = f()
-            if howlong < 0.0:
-                raise ValueError('Timeout value out of range')
-        self.__dict__['timeout'] = howlong
-
-    def gettimeout(self):
-        return self.__dict__['timeout']
-
-    def shutdown(self, how):
-        if how == 0:  # SHUT_RD
-            self.hub.cancel_wait(self._read_event, cancel_wait_ex)
-        elif how == 1:  # SHUT_WR
-            self.hub.cancel_wait(self._write_event, cancel_wait_ex)
-        else:
-            self.hub.cancel_wait(self._read_event, cancel_wait_ex)
-            self.hub.cancel_wait(self._write_event, cancel_wait_ex)
-        self._sock.shutdown(how)
 
     # sendfile: new in 3.5. But there's no real reason to not
     # support it everywhere. Note that we can't use os.sendfile()
@@ -727,6 +544,77 @@ if hasattr(_socket.socket, "share"):
     __implements__.append('fromshare')
 
 
+def _fallback_socketpair(family=AF_INET, type=SOCK_STREAM, proto=0):
+    # We originally used https://gist.github.com/4325783, by Geert Jansen. (Public domain.)
+    # We took it from 3.6 release, confirmed unchanged in 3.7 and
+    # 3.8a1. Expected to be used only on Win. Added to Win/3.5.
+    # It is always available as `socket._fallback_socketpair` from at least 3.9,
+    # We would like to stop carrying around our own implementation, but
+    # using _fallback_socketpair directly would only work if we are monkey patched.
+
+    # Current version taken from 3.13rc2
+
+    # PyPy doesn't name its fallback `_fallback_socketpair`, it uses
+    # an older copy of socket.py.
+    _LOCALHOST = '127.0.0.1'
+    _LOCALHOST_V6 = '::1'
+
+    if family == AF_INET:
+        host = _LOCALHOST
+    elif family == AF_INET6:
+        host = _LOCALHOST_V6
+    else:
+        raise ValueError("Only AF_INET and AF_INET6 socket address families "
+                         "are supported")
+    if type != SOCK_STREAM:
+        raise ValueError("Only SOCK_STREAM socket type is supported")
+    if proto != 0:
+        raise ValueError("Only protocol zero is supported")
+
+    # We create a connected TCP socket. Note the trick with
+    # setblocking(False) that prevents us from having to create a thread.
+    lsock = socket(family, type, proto)
+    try:
+        lsock.bind((host, 0))
+        lsock.listen()
+        # On IPv6, ignore flow_info and scope_id
+        addr, port = lsock.getsockname()[:2]
+        csock = socket(family, type, proto)
+        try:
+            csock.setblocking(False)
+            try:
+                csock.connect((addr, port))
+            except (BlockingIOError, InterruptedError):
+                pass
+            csock.setblocking(True)
+            ssock, _ = lsock.accept()
+        except:
+            csock.close()
+            raise
+    finally:
+        lsock.close()
+
+    # Authenticating avoids using a connection from something else
+    # able to connect to {host}:{port} instead of us.
+    # We expect only AF_INET and AF_INET6 families.
+    try:
+        if (
+            ssock.getsockname() != csock.getpeername()
+            or csock.getsockname() != ssock.getpeername()
+        ):
+            raise ConnectionError("Unexpected peer connection")
+    except:
+        # getsockname() and getpeername() can fail
+        # if either socket isn't connected.
+        ssock.close()
+        csock.close()
+        raise
+
+    return (ssock, csock)
+
+if hasattr(__socket__, _fallback_socketpair.__name__):
+    __implements__.append(_fallback_socketpair.__name__)
+
 if hasattr(_socket, "socketpair"):
 
     def socketpair(family=None, type=SOCK_STREAM, proto=0):
@@ -752,54 +640,24 @@ if hasattr(_socket, "socketpair"):
         return a, b
 
 else: # pragma: no cover
-    # Origin: https://gist.github.com/4325783, by Geert Jansen.  Public domain.
-
-    # gevent: taken from 3.6 release, confirmed unchanged in 3.7 and
-    # 3.8a1. Expected to be used only on Win. Added to Win/3.5
-
-    _LOCALHOST = '127.0.0.1'
-    _LOCALHOST_V6 = '::1'
-
-    def socketpair(family=AF_INET, type=SOCK_STREAM, proto=0):
-        if family == AF_INET:
-            host = _LOCALHOST
-        elif family == AF_INET6:
-            host = _LOCALHOST_V6
-        else:
-            raise ValueError("Only AF_INET and AF_INET6 socket address families "
-                             "are supported")
-        if type != SOCK_STREAM:
-            raise ValueError("Only SOCK_STREAM socket type is supported")
-        if proto != 0:
-            raise ValueError("Only protocol zero is supported")
-
-        # We create a connected TCP socket. Note the trick with
-        # setblocking(False) that prevents us from having to create a thread.
-        lsock = socket(family, type, proto)
-        try:
-            lsock.bind((host, 0))
-            lsock.listen()
-            # On IPv6, ignore flow_info and scope_id
-            addr, port = lsock.getsockname()[:2]
-            csock = socket(family, type, proto)
-            try:
-                csock.setblocking(False)
-                try:
-                    csock.connect((addr, port))
-                except (BlockingIOError, InterruptedError):
-                    pass
-                csock.setblocking(True)
-                ssock, _ = lsock.accept()
-            except:
-                csock.close()
-                raise
-        finally:
-            lsock.close()
-        return (ssock, csock)
+    socketpair = _fallback_socketpair
 
 
-if hasattr(__socket__, 'close'): # Python 3.7b1+
-    close = __socket__.close # pylint:disable=no-member
-    __imports__ += ['close']
 
 __all__ = __implements__ + __extensions__ + __imports__
+if _fallback_socketpair.__name__ in __all__:
+    __all__.remove(_fallback_socketpair.__name__)
+
+__version_specific__ = (
+    # Python 3.7b1+
+    'close',
+    # Python 3.10rc1+
+    'TCP_KEEPALIVE',
+    'TCP_KEEPCNT',
+)
+for _x in __version_specific__:
+    if hasattr(__socket__, _x):
+        vars()[_x] = getattr(__socket__, _x)
+        if _x not in __all__:
+            __all__.append(_x)
+del _x
